"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.parseTokens = void 0;
const childNodes_1 = require("../operators/childNodes");
const getAttribute_1 = require("../operators/getAttribute");
const hasAttribute_1 = require("../operators/hasAttribute");
const indexOf_1 = require("../operators/indexOf");
const indexOfType_1 = require("../operators/indexOfType");
const lastIndexOf_1 = require("../operators/lastIndexOf");
const lastIndexOfType_1 = require("../operators/lastIndexOfType");
const parents_1 = require("../operators/parents");
const previousSibling_1 = require("../operators/previousSibling");
const textContent_1 = require("../operators/textContent");
const error_1 = require("./error");
const unescape_1 = require("./unescape");
const tokens_1 = require("./tokens");
function parseTokens(tokens) {
    const groupedCheckers = [];
    let chainedCheckers = [];
    let checkers = [];
    const length = tokens.length + 1;
    let cursor = 0;
    const skip = (...types) => {
        var _a;
        if (types.includes((_a = tokens[cursor]) === null || _a === void 0 ? void 0 : _a.type)) {
            ++cursor;
            return true;
        }
        return false;
    };
    const is = (...types) => {
        const next = tokens[cursor];
        return next && types.includes(next.type) ? next : null;
    };
    const not = (...types) => {
        const next = tokens[cursor];
        return next && !types.includes(next.type) ? next : null;
    };
    const nextName = (...extraTokens) => {
        let start = null;
        let end = null;
        let text = '';
        let next = is(tokens_1.TOKEN_IDENT, tokens_1.TOKEN_MINUS, ...extraTokens);
        while (next) {
            if (start == null)
                start = next.start;
            end = next.end;
            text += next.text;
            ++cursor;
            next = is(tokens_1.TOKEN_IDENT, tokens_1.TOKEN_MINUS, ...extraTokens);
        }
        return text.length === 0 || start == null || end == null
            ? null
            : {
                type: tokens_1.TOKEN_IDENT,
                text,
                start,
                end,
            };
    };
    while (cursor < length) {
        const token = tokens[cursor++];
        if (token == null || token.type === tokens_1.TOKEN_COMMA) {
            if (checkers.length > 0) {
                const grouped = groupCheckers(checkers);
                if (grouped) {
                    chainedCheckers.unshift([tokens_1.TOKEN_SPACE, grouped]);
                    checkers = [];
                }
            }
            const chained = chainCheckers(chainedCheckers);
            if (chained) {
                groupedCheckers.push(chained);
                chainedCheckers = [];
                continue;
            }
        }
        else if (token.type === tokens_1.TOKEN_SPACE || token.type === tokens_1.TOKEN_GT || token.type === tokens_1.TOKEN_PLUS || token.type === tokens_1.TOKEN_TILDE) {
            const grouped = groupCheckers(checkers);
            if (grouped) {
                chainedCheckers.unshift([token.type, grouped]);
                checkers = [];
                continue;
            }
            if (token.type === tokens_1.TOKEN_SPACE) {
                continue;
            }
        }
        else if (token.type === tokens_1.TOKEN_COLON) {
            const extraColon = is(tokens_1.TOKEN_COLON) ? token : null;
            if (extraColon)
                cursor++;
            const nameToken = nextName();
            if (nameToken) {
                const pseudoName = nameToken.text;
                if (pseudoName === 'root') {
                    checkers.push((node) => node.parent == null || node.parent.tagName === '');
                    continue;
                }
                else if (pseudoName === 'scope') {
                    checkers.push((node, scope) => node === scope);
                    continue;
                }
                else if (pseudoName === 'empty') {
                    checkers.push((node) => (0, childNodes_1.childNodes)(node).length === 0 && (0, textContent_1.textContent)(node).length === 0);
                    continue;
                }
                else if (pseudoName === 'disabled') {
                    checkers.push((node) => { var _a; return ['disabled', 'true'].includes((_a = (0, getAttribute_1.getAttribute)(node, 'disabled')) !== null && _a !== void 0 ? _a : ''); });
                    continue;
                }
                else if (pseudoName === 'enabled') {
                    checkers.push((node) => { var _a; return !['disabled', 'true'].includes((_a = (0, getAttribute_1.getAttribute)(node, 'disabled')) !== null && _a !== void 0 ? _a : ''); });
                    continue;
                }
                else if (pseudoName === 'first-child') {
                    checkers.push((node) => (0, indexOf_1.indexOf)(node) === 0);
                    continue;
                }
                else if (pseudoName === 'last-child') {
                    checkers.push((node) => (0, lastIndexOf_1.lastIndexOf)(node) === 0);
                    continue;
                }
                else if (pseudoName === 'only-child') {
                    checkers.push((node) => node.parent == null || (0, childNodes_1.childNodes)(node.parent).length === 1);
                    continue;
                }
                else if (pseudoName === 'first-of-type') {
                    checkers.push((node) => (0, indexOfType_1.indexOfType)(node) === 0);
                    continue;
                }
                else if (pseudoName === 'last-of-type') {
                    checkers.push((node) => (0, lastIndexOfType_1.lastIndexOfType)(node) === 0);
                    continue;
                }
                else if (pseudoName === 'only-of-type') {
                    checkers.push((node) => (0, indexOfType_1.indexOfType)(node) === 0 && (0, lastIndexOfType_1.lastIndexOfType)(node) === 0);
                    continue;
                }
                else if (['first-letter', 'first-line', 'before', 'after'].includes(pseudoName)) {
                    checkers.push(() => false);
                    continue;
                }
                const pseudoParam = [];
                skip(tokens_1.TOKEN_ROUND_OPEN);
                let paramToken;
                while ((paramToken = not(tokens_1.TOKEN_ROUND_CLOSE))) {
                    ++cursor;
                    pseudoParam.push(paramToken);
                }
                skip(tokens_1.TOKEN_ROUND_CLOSE);
                if (pseudoName === 'nth-child') {
                    const expr = parseExpression(pseudoParam);
                    if (expr) {
                        checkers.push((node) => expr((0, indexOf_1.indexOf)(node) + 1));
                        continue;
                    }
                }
                else if (pseudoName === 'nth-last-child') {
                    const expr = parseExpression(pseudoParam);
                    if (expr) {
                        checkers.push((node) => expr((0, lastIndexOf_1.lastIndexOf)(node) + 1));
                        continue;
                    }
                }
                else if (pseudoName === 'nth-of-type') {
                    const expr = parseExpression(pseudoParam);
                    if (expr) {
                        checkers.push((node) => expr((0, indexOfType_1.indexOfType)(node) + 1));
                        continue;
                    }
                }
                else if (pseudoName === 'nth-last-of-type') {
                    const expr = parseExpression(pseudoParam);
                    if (expr) {
                        checkers.push((node) => expr((0, lastIndexOfType_1.lastIndexOfType)(node) + 1));
                        continue;
                    }
                }
                else if (pseudoName === 'not') {
                    const notChecker = parseTokens(pseudoParam);
                    if (notChecker) {
                        checkers.push((node, scope) => !notChecker(node, scope));
                        continue;
                    }
                }
                else if (['slotted'].includes(pseudoName)) {
                    checkers.push(() => false);
                    continue;
                }
                else {
                    throw (0, error_1.createSyntaxError)(extraColon !== null && extraColon !== void 0 ? extraColon : nameToken);
                }
            }
        }
        else if (token.type === tokens_1.TOKEN_SQUARE_OPEN) {
            const nameToken = nextName(tokens_1.TOKEN_COLON);
            if (nameToken) {
                const attrName = nameToken.text;
                let attrValue = null;
                const cmpToken = not(tokens_1.TOKEN_SQUARE_CLOSE);
                if (cmpToken) {
                    ++cursor;
                    skip(tokens_1.TOKEN_QUOTE);
                    const valueToken = is(tokens_1.TOKEN_STRING, tokens_1.TOKEN_IDENT);
                    if (valueToken) {
                        ++cursor;
                        attrValue = valueToken.text;
                    }
                    else
                        attrValue = '';
                    skip(tokens_1.TOKEN_QUOTE);
                }
                if (attrValue == null) {
                    checkers.push((node) => (0, hasAttribute_1.hasAttribute)(node, attrName));
                }
                else {
                    let attrValueUnescape = (0, unescape_1.unescape)(attrValue);
                    if (cmpToken != null) {
                        if (cmpToken.type === tokens_1.TOKEN_EQUAL) {
                            checkers.push((node) => (0, getAttribute_1.getAttribute)(node, attrName) === attrValueUnescape);
                        }
                        else if (cmpToken.type === tokens_1.TOKEN_CARET_EQUAL) {
                            checkers.push((node) => { var _a; return attrValueUnescape !== '' && !!((_a = (0, getAttribute_1.getAttribute)(node, attrName)) === null || _a === void 0 ? void 0 : _a.startsWith(attrValueUnescape)); });
                        }
                        else if (cmpToken.type === tokens_1.TOKEN_DOLLAR_EQUAL) {
                            checkers.push((node) => { var _a; return attrValueUnescape !== '' && !!((_a = (0, getAttribute_1.getAttribute)(node, attrName)) === null || _a === void 0 ? void 0 : _a.endsWith(attrValueUnescape)); });
                        }
                        else if (cmpToken.type === tokens_1.TOKEN_ASTERISK_EQUAL) {
                            attrValueUnescape = attrValueUnescape.trim();
                            checkers.push((node) => { var _a; return attrValueUnescape !== '' && !!((_a = (0, getAttribute_1.getAttribute)(node, attrName)) === null || _a === void 0 ? void 0 : _a.includes(attrValueUnescape)); });
                        }
                        else if (cmpToken.type === tokens_1.TOKEN_TILDE_EQUAL) {
                            checkers.push((node) => {
                                if (attrValueUnescape === '')
                                    return false;
                                const list = (0, getAttribute_1.getAttribute)(node, attrName, '').split(' ');
                                return list.length > 0 && list.includes(attrValueUnescape);
                            });
                        }
                        else if (cmpToken.type === tokens_1.TOKEN_PIPE_EQUAL) {
                            checkers.push((node) => {
                                if (attrValueUnescape === '')
                                    return false;
                                const list = (0, getAttribute_1.getAttribute)(node, attrName, '').split('-');
                                return list.length > 0 && list.includes(attrValueUnescape);
                            });
                        }
                        else
                            throw (0, error_1.createSyntaxError)(cmpToken);
                    }
                }
                if (skip(tokens_1.TOKEN_SQUARE_CLOSE)) {
                    continue;
                }
                if (tokens[cursor + 1] == null)
                    continue;
            }
        }
        else if (token.type === tokens_1.TOKEN_IDENT) {
            --cursor;
            const nameToken = nextName();
            if (nameToken) {
                const nodeName = (0, unescape_1.unescape)(nameToken.text);
                checkers.push((node) => node.tagName === nodeName);
                continue;
            }
        }
        else if (token.type === tokens_1.TOKEN_HASH) {
            const nameToken = nextName(tokens_1.TOKEN_NUMBER);
            if (nameToken) {
                const nodeName = (0, unescape_1.unescape)(nameToken.text);
                checkers.push((node) => (0, getAttribute_1.getAttribute)(node, 'id') === nodeName);
                continue;
            }
        }
        else if (token.type === tokens_1.TOKEN_DOT) {
            const nameToken = nextName();
            if (nameToken) {
                const nodeName = (0, unescape_1.unescape)(nameToken.text);
                checkers.push((node) => (0, getAttribute_1.getAttribute)(node, 'class', '').split(' ').includes(nodeName));
                continue;
            }
        }
        else if (token.type === tokens_1.TOKEN_ASTERISK) {
            if (checkers.length === 0) {
                checkers.unshift((node) => node.tagName !== '');
            }
            continue;
        }
        throw (0, error_1.createSyntaxError)(token);
    }
    if (groupedCheckers.length === 0) {
        throw new SyntaxError(`The provided selector is empty.`);
    }
    return (node, scope) => groupedCheckers.some((check) => check(node, scope));
}
exports.parseTokens = parseTokens;
function groupCheckers(checkers) {
    if (checkers.length) {
        return (node, scope) => checkers.every((check) => check(node, scope));
    }
    return null;
}
function chainCheckers(checkers) {
    if (checkers.length === 0)
        return null;
    if (checkers.length === 1)
        return checkers[0][1];
    return (node, scope) => {
        if (!checkers[0][1](node, scope))
            return false;
        for (let index = 1, parent = node; parent != null; index++) {
            const [type, check] = checkers[index];
            if (parent.parent) {
                if (type === tokens_1.TOKEN_SPACE) {
                    parent = (0, parents_1.parents)(parent, true).find((node2) => check(node2, scope));
                }
                else if (type === tokens_1.TOKEN_GT) {
                    parent = check(parent.parent, scope) ? parent.parent : undefined;
                }
                else if (type === tokens_1.TOKEN_PLUS) {
                    const prev = (0, previousSibling_1.previousSibling)(parent);
                    parent = prev && check(prev, scope) ? prev : undefined;
                }
                else if (type === tokens_1.TOKEN_TILDE) {
                    const nodeIndex = (0, indexOf_1.indexOf)(parent);
                    if (nodeIndex === 0)
                        parent = undefined;
                    else {
                        const precursors = (0, childNodes_1.childNodes)(parent.parent).slice(0, nodeIndex);
                        parent = precursors.find((node2) => check(node2, scope));
                    }
                }
            }
            else
                parent = undefined;
            if (parent != null && index === checkers.length - 1)
                return true;
        }
        return false;
    };
}
function parseExpression(expr) {
    if (expr.length === 0)
        return null;
    if (expr.length <= 5) {
        let pos = 0;
        let multiplier = 1;
        let offset = 0;
        const firstSign = expr[pos] && (expr[pos].type === tokens_1.TOKEN_PLUS || expr[pos].type === tokens_1.TOKEN_MINUS) ? expr[pos].type : null;
        if (firstSign)
            pos++;
        let firstNumber = undefined;
        if (expr[pos] && expr[pos].type === tokens_1.TOKEN_NUMBER) {
            firstNumber = parseInt(expr[pos].text, 10);
            pos++;
        }
        const hasVariable = expr[pos] && expr[pos].type === tokens_1.TOKEN_IDENT;
        if (hasVariable) {
            if (expr[pos].text !== 'n')
                throw (0, error_1.createSyntaxError)(expr[pos]);
            pos++;
        }
        const secondSign = expr[pos] && (expr[pos].type === tokens_1.TOKEN_PLUS || expr[pos].type === tokens_1.TOKEN_MINUS) ? expr[pos].type : null;
        if (secondSign)
            pos++;
        let secondNumber = undefined;
        if (secondSign != null) {
            if (expr[pos] && expr[pos].type === tokens_1.TOKEN_NUMBER) {
                secondNumber = parseInt(expr[pos].text, 10);
                pos++;
            }
            else {
                throw (0, error_1.createSyntaxError)(expr[pos]);
            }
        }
        if (pos !== expr.length)
            throw (0, error_1.createSyntaxError)(expr[pos]);
        if (hasVariable) {
            multiplier = firstNumber !== null && firstNumber !== void 0 ? firstNumber : 1;
        }
        if (secondSign != null) {
            if (secondSign === tokens_1.TOKEN_PLUS) {
                offset = secondNumber !== null && secondNumber !== void 0 ? secondNumber : 0;
            }
            else if (secondSign === tokens_1.TOKEN_MINUS) {
                offset = -(secondNumber !== null && secondNumber !== void 0 ? secondNumber : 0);
            }
        }
        if (hasVariable) {
            if (firstSign === tokens_1.TOKEN_MINUS) {
                return (index) => {
                    index -= offset;
                    return index <= 0 && index % multiplier === 0;
                };
            }
            else {
                return (index) => {
                    index -= offset;
                    return index >= 0 && index % multiplier === 0;
                };
            }
        }
        else {
            return (index) => index === firstNumber;
        }
    }
    throw (0, error_1.createSyntaxError)(expr[5]);
}
//# sourceMappingURL=parser.js.map