var __read = (this && this.__read) || function (o, n) {
    var m = typeof Symbol === "function" && o[Symbol.iterator];
    if (!m) return o;
    var i = m.call(o), r, ar = [], e;
    try {
        while ((n === void 0 || n-- > 0) && !(r = i.next()).done) ar.push(r.value);
    }
    catch (error) { e = { error: error }; }
    finally {
        try {
            if (r && !r.done && (m = i["return"])) m.call(i);
        }
        finally { if (e) throw e.error; }
    }
    return ar;
};
var __spreadArray = (this && this.__spreadArray) || function (to, from, pack) {
    if (pack || arguments.length === 2) for (var i = 0, l = from.length, ar; i < l; i++) {
        if (ar || !(i in from)) {
            if (!ar) ar = Array.prototype.slice.call(from, 0, i);
            ar[i] = from[i];
        }
    }
    return to.concat(ar || Array.prototype.slice.call(from));
};
import { childNodes } from '../operators/childNodes';
import { getAttribute } from '../operators/getAttribute';
import { hasAttribute } from '../operators/hasAttribute';
import { indexOf } from '../operators/indexOf';
import { indexOfType } from '../operators/indexOfType';
import { lastIndexOf } from '../operators/lastIndexOf';
import { lastIndexOfType } from '../operators/lastIndexOfType';
import { parents } from '../operators/parents';
import { previousSibling } from '../operators/previousSibling';
import { textContent } from '../operators/textContent';
import { createSyntaxError } from './error';
import { unescape } from './unescape';
import { TOKEN_SPACE, TOKEN_GT, TOKEN_PLUS, TOKEN_TILDE, TOKEN_IDENT, TOKEN_MINUS, TOKEN_COMMA, TOKEN_COLON, TOKEN_ROUND_OPEN, TOKEN_ROUND_CLOSE, TOKEN_SQUARE_OPEN, TOKEN_SQUARE_CLOSE, TOKEN_QUOTE, TOKEN_STRING, TOKEN_EQUAL, TOKEN_CARET_EQUAL, TOKEN_DOLLAR_EQUAL, TOKEN_ASTERISK_EQUAL, TOKEN_TILDE_EQUAL, TOKEN_PIPE_EQUAL, TOKEN_HASH, TOKEN_NUMBER, TOKEN_DOT, TOKEN_ASTERISK, } from './tokens';
export function parseTokens(tokens) {
    var groupedCheckers = [];
    var chainedCheckers = [];
    var checkers = [];
    var length = tokens.length + 1;
    var cursor = 0;
    var skip = function () {
        var _a;
        var types = [];
        for (var _i = 0; _i < arguments.length; _i++) {
            types[_i] = arguments[_i];
        }
        if (types.includes((_a = tokens[cursor]) === null || _a === void 0 ? void 0 : _a.type)) {
            ++cursor;
            return true;
        }
        return false;
    };
    var is = function () {
        var types = [];
        for (var _i = 0; _i < arguments.length; _i++) {
            types[_i] = arguments[_i];
        }
        var next = tokens[cursor];
        return next && types.includes(next.type) ? next : null;
    };
    var not = function () {
        var types = [];
        for (var _i = 0; _i < arguments.length; _i++) {
            types[_i] = arguments[_i];
        }
        var next = tokens[cursor];
        return next && !types.includes(next.type) ? next : null;
    };
    var nextName = function () {
        var extraTokens = [];
        for (var _i = 0; _i < arguments.length; _i++) {
            extraTokens[_i] = arguments[_i];
        }
        var start = null;
        var end = null;
        var text = '';
        var next = is.apply(void 0, __spreadArray([TOKEN_IDENT, TOKEN_MINUS], __read(extraTokens), false));
        while (next) {
            if (start == null)
                start = next.start;
            end = next.end;
            text += next.text;
            ++cursor;
            next = is.apply(void 0, __spreadArray([TOKEN_IDENT, TOKEN_MINUS], __read(extraTokens), false));
        }
        return text.length === 0 || start == null || end == null
            ? null
            : {
                type: TOKEN_IDENT,
                text: text,
                start: start,
                end: end,
            };
    };
    var _loop_1 = function () {
        var token = tokens[cursor++];
        if (token == null || token.type === TOKEN_COMMA) {
            if (checkers.length > 0) {
                var grouped = groupCheckers(checkers);
                if (grouped) {
                    chainedCheckers.unshift([TOKEN_SPACE, grouped]);
                    checkers = [];
                }
            }
            var chained = chainCheckers(chainedCheckers);
            if (chained) {
                groupedCheckers.push(chained);
                chainedCheckers = [];
                return "continue";
            }
        }
        else if (token.type === TOKEN_SPACE || token.type === TOKEN_GT || token.type === TOKEN_PLUS || token.type === TOKEN_TILDE) {
            var grouped = groupCheckers(checkers);
            if (grouped) {
                chainedCheckers.unshift([token.type, grouped]);
                checkers = [];
                return "continue";
            }
            if (token.type === TOKEN_SPACE) {
                return "continue";
            }
        }
        else if (token.type === TOKEN_COLON) {
            var extraColon = is(TOKEN_COLON) ? token : null;
            if (extraColon)
                cursor++;
            var nameToken = nextName();
            if (nameToken) {
                var pseudoName = nameToken.text;
                if (pseudoName === 'root') {
                    checkers.push(function (node) { return node.parent == null || node.parent.tagName === ''; });
                    return "continue";
                }
                else if (pseudoName === 'scope') {
                    checkers.push(function (node, scope) { return node === scope; });
                    return "continue";
                }
                else if (pseudoName === 'empty') {
                    checkers.push(function (node) { return childNodes(node).length === 0 && textContent(node).length === 0; });
                    return "continue";
                }
                else if (pseudoName === 'disabled') {
                    checkers.push(function (node) { var _a; return ['disabled', 'true'].includes((_a = getAttribute(node, 'disabled')) !== null && _a !== void 0 ? _a : ''); });
                    return "continue";
                }
                else if (pseudoName === 'enabled') {
                    checkers.push(function (node) { var _a; return !['disabled', 'true'].includes((_a = getAttribute(node, 'disabled')) !== null && _a !== void 0 ? _a : ''); });
                    return "continue";
                }
                else if (pseudoName === 'first-child') {
                    checkers.push(function (node) { return indexOf(node) === 0; });
                    return "continue";
                }
                else if (pseudoName === 'last-child') {
                    checkers.push(function (node) { return lastIndexOf(node) === 0; });
                    return "continue";
                }
                else if (pseudoName === 'only-child') {
                    checkers.push(function (node) { return node.parent == null || childNodes(node.parent).length === 1; });
                    return "continue";
                }
                else if (pseudoName === 'first-of-type') {
                    checkers.push(function (node) { return indexOfType(node) === 0; });
                    return "continue";
                }
                else if (pseudoName === 'last-of-type') {
                    checkers.push(function (node) { return lastIndexOfType(node) === 0; });
                    return "continue";
                }
                else if (pseudoName === 'only-of-type') {
                    checkers.push(function (node) { return indexOfType(node) === 0 && lastIndexOfType(node) === 0; });
                    return "continue";
                }
                else if (['first-letter', 'first-line', 'before', 'after'].includes(pseudoName)) {
                    checkers.push(function () { return false; });
                    return "continue";
                }
                var pseudoParam = [];
                skip(TOKEN_ROUND_OPEN);
                var paramToken = void 0;
                while ((paramToken = not(TOKEN_ROUND_CLOSE))) {
                    ++cursor;
                    pseudoParam.push(paramToken);
                }
                skip(TOKEN_ROUND_CLOSE);
                if (pseudoName === 'nth-child') {
                    var expr_1 = parseExpression(pseudoParam);
                    if (expr_1) {
                        checkers.push(function (node) { return expr_1(indexOf(node) + 1); });
                        return "continue";
                    }
                }
                else if (pseudoName === 'nth-last-child') {
                    var expr_2 = parseExpression(pseudoParam);
                    if (expr_2) {
                        checkers.push(function (node) { return expr_2(lastIndexOf(node) + 1); });
                        return "continue";
                    }
                }
                else if (pseudoName === 'nth-of-type') {
                    var expr_3 = parseExpression(pseudoParam);
                    if (expr_3) {
                        checkers.push(function (node) { return expr_3(indexOfType(node) + 1); });
                        return "continue";
                    }
                }
                else if (pseudoName === 'nth-last-of-type') {
                    var expr_4 = parseExpression(pseudoParam);
                    if (expr_4) {
                        checkers.push(function (node) { return expr_4(lastIndexOfType(node) + 1); });
                        return "continue";
                    }
                }
                else if (pseudoName === 'not') {
                    var notChecker_1 = parseTokens(pseudoParam);
                    if (notChecker_1) {
                        checkers.push(function (node, scope) { return !notChecker_1(node, scope); });
                        return "continue";
                    }
                }
                else if (['slotted'].includes(pseudoName)) {
                    checkers.push(function () { return false; });
                    return "continue";
                }
                else {
                    throw createSyntaxError(extraColon !== null && extraColon !== void 0 ? extraColon : nameToken);
                }
            }
        }
        else if (token.type === TOKEN_SQUARE_OPEN) {
            var nameToken = nextName(TOKEN_COLON);
            if (nameToken) {
                var attrName_1 = nameToken.text;
                var attrValue = null;
                var cmpToken = not(TOKEN_SQUARE_CLOSE);
                if (cmpToken) {
                    ++cursor;
                    skip(TOKEN_QUOTE);
                    var valueToken = is(TOKEN_STRING, TOKEN_IDENT);
                    if (valueToken) {
                        ++cursor;
                        attrValue = valueToken.text;
                    }
                    else
                        attrValue = '';
                    skip(TOKEN_QUOTE);
                }
                if (attrValue == null) {
                    checkers.push(function (node) { return hasAttribute(node, attrName_1); });
                }
                else {
                    var attrValueUnescape_1 = unescape(attrValue);
                    if (cmpToken != null) {
                        if (cmpToken.type === TOKEN_EQUAL) {
                            checkers.push(function (node) { return getAttribute(node, attrName_1) === attrValueUnescape_1; });
                        }
                        else if (cmpToken.type === TOKEN_CARET_EQUAL) {
                            checkers.push(function (node) { var _a; return attrValueUnescape_1 !== '' && !!((_a = getAttribute(node, attrName_1)) === null || _a === void 0 ? void 0 : _a.startsWith(attrValueUnescape_1)); });
                        }
                        else if (cmpToken.type === TOKEN_DOLLAR_EQUAL) {
                            checkers.push(function (node) { var _a; return attrValueUnescape_1 !== '' && !!((_a = getAttribute(node, attrName_1)) === null || _a === void 0 ? void 0 : _a.endsWith(attrValueUnescape_1)); });
                        }
                        else if (cmpToken.type === TOKEN_ASTERISK_EQUAL) {
                            attrValueUnescape_1 = attrValueUnescape_1.trim();
                            checkers.push(function (node) { var _a; return attrValueUnescape_1 !== '' && !!((_a = getAttribute(node, attrName_1)) === null || _a === void 0 ? void 0 : _a.includes(attrValueUnescape_1)); });
                        }
                        else if (cmpToken.type === TOKEN_TILDE_EQUAL) {
                            checkers.push(function (node) {
                                if (attrValueUnescape_1 === '')
                                    return false;
                                var list = getAttribute(node, attrName_1, '').split(' ');
                                return list.length > 0 && list.includes(attrValueUnescape_1);
                            });
                        }
                        else if (cmpToken.type === TOKEN_PIPE_EQUAL) {
                            checkers.push(function (node) {
                                if (attrValueUnescape_1 === '')
                                    return false;
                                var list = getAttribute(node, attrName_1, '').split('-');
                                return list.length > 0 && list.includes(attrValueUnescape_1);
                            });
                        }
                        else
                            throw createSyntaxError(cmpToken);
                    }
                }
                if (skip(TOKEN_SQUARE_CLOSE)) {
                    return "continue";
                }
                if (tokens[cursor + 1] == null)
                    return "continue";
            }
        }
        else if (token.type === TOKEN_IDENT) {
            --cursor;
            var nameToken = nextName();
            if (nameToken) {
                var nodeName_1 = unescape(nameToken.text);
                checkers.push(function (node) { return node.tagName === nodeName_1; });
                return "continue";
            }
        }
        else if (token.type === TOKEN_HASH) {
            var nameToken = nextName(TOKEN_NUMBER);
            if (nameToken) {
                var nodeName_2 = unescape(nameToken.text);
                checkers.push(function (node) { return getAttribute(node, 'id') === nodeName_2; });
                return "continue";
            }
        }
        else if (token.type === TOKEN_DOT) {
            var nameToken = nextName();
            if (nameToken) {
                var nodeName_3 = unescape(nameToken.text);
                checkers.push(function (node) { return getAttribute(node, 'class', '').split(' ').includes(nodeName_3); });
                return "continue";
            }
        }
        else if (token.type === TOKEN_ASTERISK) {
            if (checkers.length === 0) {
                checkers.unshift(function (node) { return node.tagName !== ''; });
            }
            return "continue";
        }
        throw createSyntaxError(token);
    };
    while (cursor < length) {
        _loop_1();
    }
    if (groupedCheckers.length === 0) {
        throw new SyntaxError("The provided selector is empty.");
    }
    return function (node, scope) { return groupedCheckers.some(function (check) { return check(node, scope); }); };
}
function groupCheckers(checkers) {
    if (checkers.length) {
        return function (node, scope) { return checkers.every(function (check) { return check(node, scope); }); };
    }
    return null;
}
function chainCheckers(checkers) {
    if (checkers.length === 0)
        return null;
    if (checkers.length === 1)
        return checkers[0][1];
    return function (node, scope) {
        if (!checkers[0][1](node, scope))
            return false;
        var _loop_2 = function (index, parent) {
            var _a = __read(checkers[index], 2), type = _a[0], check = _a[1];
            if (parent.parent) {
                if (type === TOKEN_SPACE) {
                    parent = parents(parent, true).find(function (node2) { return check(node2, scope); });
                }
                else if (type === TOKEN_GT) {
                    parent = check(parent.parent, scope) ? parent.parent : undefined;
                }
                else if (type === TOKEN_PLUS) {
                    var prev = previousSibling(parent);
                    parent = prev && check(prev, scope) ? prev : undefined;
                }
                else if (type === TOKEN_TILDE) {
                    var nodeIndex = indexOf(parent);
                    if (nodeIndex === 0)
                        parent = undefined;
                    else {
                        var precursors = childNodes(parent.parent).slice(0, nodeIndex);
                        parent = precursors.find(function (node2) { return check(node2, scope); });
                    }
                }
            }
            else
                parent = undefined;
            if (parent != null && index === checkers.length - 1)
                return { value: true };
            out_parent_1 = parent;
        };
        var out_parent_1;
        for (var index = 1, parent = node; parent != null; index++) {
            var state_1 = _loop_2(index, parent);
            parent = out_parent_1;
            if (typeof state_1 === "object")
                return state_1.value;
        }
        return false;
    };
}
function parseExpression(expr) {
    if (expr.length === 0)
        return null;
    if (expr.length <= 5) {
        var pos = 0;
        var multiplier_1 = 1;
        var offset_1 = 0;
        var firstSign = expr[pos] && (expr[pos].type === TOKEN_PLUS || expr[pos].type === TOKEN_MINUS) ? expr[pos].type : null;
        if (firstSign)
            pos++;
        var firstNumber_1 = undefined;
        if (expr[pos] && expr[pos].type === TOKEN_NUMBER) {
            firstNumber_1 = parseInt(expr[pos].text, 10);
            pos++;
        }
        var hasVariable = expr[pos] && expr[pos].type === TOKEN_IDENT;
        if (hasVariable) {
            if (expr[pos].text !== 'n')
                throw createSyntaxError(expr[pos]);
            pos++;
        }
        var secondSign = expr[pos] && (expr[pos].type === TOKEN_PLUS || expr[pos].type === TOKEN_MINUS) ? expr[pos].type : null;
        if (secondSign)
            pos++;
        var secondNumber = undefined;
        if (secondSign != null) {
            if (expr[pos] && expr[pos].type === TOKEN_NUMBER) {
                secondNumber = parseInt(expr[pos].text, 10);
                pos++;
            }
            else {
                throw createSyntaxError(expr[pos]);
            }
        }
        if (pos !== expr.length)
            throw createSyntaxError(expr[pos]);
        if (hasVariable) {
            multiplier_1 = firstNumber_1 !== null && firstNumber_1 !== void 0 ? firstNumber_1 : 1;
        }
        if (secondSign != null) {
            if (secondSign === TOKEN_PLUS) {
                offset_1 = secondNumber !== null && secondNumber !== void 0 ? secondNumber : 0;
            }
            else if (secondSign === TOKEN_MINUS) {
                offset_1 = -(secondNumber !== null && secondNumber !== void 0 ? secondNumber : 0);
            }
        }
        if (hasVariable) {
            if (firstSign === TOKEN_MINUS) {
                return function (index) {
                    index -= offset_1;
                    return index <= 0 && index % multiplier_1 === 0;
                };
            }
            else {
                return function (index) {
                    index -= offset_1;
                    return index >= 0 && index % multiplier_1 === 0;
                };
            }
        }
        else {
            return function (index) { return index === firstNumber_1; };
        }
    }
    throw createSyntaxError(expr[5]);
}
//# sourceMappingURL=parser.js.map